class Filters:
    def __init__(self, qp=None):
        self.qp = qp
        self.init = {}
        self.equals = {}
        self.personalize = []
        self.customize = {}

    def process(self):
        for k, v in self.qp.items():
            if k.find('filter') >= 0:
                attr = k[k.find('[') + 1 : k.find(']')]
                t = self.process_value(attr, v)
                if attr in self.personalize:
                    self.customize = {**self.customize, **t}
                else:
                    self.equals = {**self.equals, **t}

    def get(self):
        r = {'init': self.init, 'equals': self.equals, 'customize': self.customize}
        return r

    @staticmethod
    def process_value(attr, value):
        parts = value.split('|')
        if len(parts) > 1:
            value = parts

        return {attr: value}



from .datagrid.datagrid import DataGrid
from .datagrid.filters import Filters
from .datagrid.query_builder import QueryBuilder
from .datagrid.query_params import QueryParams
from .datagrid.paginator import Paginator
from .render import APIRender
from .exception import APIException



class APIException(Exception):
    def __init__(self,  message: str|None, code: int, detail: dict = None):
        self.message = message
        self.code = code
        self.detail = detail or {}

    def get_error(self) -> str|dict:
        if self.detail:
            # Check if resposne from External API attribute (response)
            if 'response' in self.detail:
                if self.detail['response']:
                    return self.detail['response']
                else:
                    return self.message

            return self.detail

        return self.message



from fastapi import Request
from sqlmodel.ext.asyncio.session import AsyncSession

from .paginator import Paginator
from .query_builder import QueryBuilder
from .query_params import QueryParams


class DataGrid:
    def __init__(self, db: AsyncSession, model, request: Request):

        self.model = model
        self.db = db
        self.params = {}

        self.qp: QueryParams = QueryParams(request)
        self.qb: QueryBuilder | None = None

    def init(self):
        self.params = self.qp.get()
        self.qb = QueryBuilder(self.db, self.model, self.params).init()

        return self

    async def get(self):
        match self.qp.view:
            case 'paginate':
                return await Paginator(self.qb, self.qp).get()
            case 'count':
                return await self.qb.total()
            case _:
                return await self.qb.get()



from fastapi import Request

from .filters import Filters


class QueryParams:
    def __init__(self, request: Request, qp=None):
        self.qp  = qp.copy() if qp else None
        self.filters = Filters(self.qp)
        self.sort = []
        self.view = 'list'
        self.request = request
        self.page = 1
        self.size = 15
        self.r = {}

    def process_sort(self) -> None:
        if 'sort' in self.qp:
            direction = 'asc'
            col = self.qp['sort']
            if self.qp['sort'].find('-', 0, 1) >= 0:
                direction = 'desc'
                col = self.qp['sort'][1:]

            self.sort = [{'col': col, 'dir': direction, 'by': self.qp['sort']}]

    def process(self) -> None:
        self.process_sort()

        if 'page' in self.qp:
            self.page = int(self.qp['page'])

        if 'size' in self.qp:
            self.size = int(self.qp['size'])

        if 'view' in self.qp:
            self.view = self.qp['view']

    def get(self):
        self.qp = dict(self.request.query_params)
        self.filters.qp = dict(self.request.query_params)

        # Filters
        self.process()
        self.filters.process()
        filters = self.filters.get()
        if any(filters):
            self.r['filters'] = filters

        # With
        parts = []
        if 'with' in self.qp:
            parts = self.qp['with'].split(',')
            for idx, x in enumerate(parts):
                if x.find('.'):
                    x = x.replace('.', '__')

                parts[idx] = x.strip()

        self.r['with'] = parts
        self.r['sort'] = self.sort
        self.r['view'] = self.view
        self.r['page'] = self.page
        self.r['size'] = self.size

        return self.r



from sqlmodel.ext.asyncio.session import AsyncSession
from sqlmodel import select
from sqlalchemy.orm import selectinload


class QueryBuilder:
    def __init__(self, db: AsyncSession, model, params):
        self.model = model
        self.stmt = None
        self.params = params
        self.db = db

    def init(self):
        filters = {}
        if ('filters' in self.params) and ('equals' in self.params['filters']):
            filters = self.params['filters']['equals']
            if 'init' in self.params['filters']:
                filters = {**filters, **self.params['filters']['init']}

        self.stmt = select(self.model)
        self.apply_filters(filters)
        self.apply_sort()

        if self.params['with']:
            relationships = self.params['with']
            self.stmt = self.stmt.options(*[selectinload(getattr(self.model, relation)) for relation in relationships])

        return self

    def apply_filters(self, filters: dict[str, any]):
        for key, value in filters.items():
            if value:
                column = getattr(self.model, key, None)
                if column:
                    if isinstance(value, list):
                        self.stmt = self.stmt.where(column.in_(value))
                    else:
                        self.stmt = self.stmt.where(column == value)

    def apply_sort(self):
        if self.params['sort']:
            col = self.params['sort'][0]['col']
            dir_ = self.params['sort'][0]['dir']

            column = getattr(self.model, col, None)
            if column:
                if dir_ == 'desc':
                    self.stmt = self.stmt.order_by(column.desc())
                else:
                    self.stmt = self.stmt.order_by(column)

    async def get(self, native = False):
        result = await self.db.exec(self.stmt)
        if native:
            return result.all()
        else:
            rows = result.all()
            # Check only the first record, if available
            if rows and (not hasattr(rows[0], 'to_dict') or not callable(getattr(rows[0], 'to_dict', None))):
                return rows

            # If the first record has `to_dict()`, return the transformed list
            return [o.to_dict() for o in rows]

    async def total(self) -> int:
        all_ = await self.get(True)
        return len(all_)



import json
from http import HTTPStatus
from typing import Any

from fastapi import Request, HTTPException
from fastapi.responses import JSONResponse
from pydantic import ValidationError

from .exception import APIException


class APIRender:
    @staticmethod
    def display(message, code=200):
        message = APIRender.clean_message(message)
        response = message if isinstance(message, dict) else {'title': message}
        response = APIRender.format_message(response, code)
        return JSONResponse(
            status_code=code,
            content=response
        )

    @staticmethod
    def format_message(message: dict, code: int):
        formatted_message = {
            'title': '',
            'type': 'about:blank',
            'detail': None,
            'status': code,
            'instance': '',
        }

        if 'title' in message:
            formatted_message['title'] = message['title']

        if not formatted_message['title'] and 'message' in message:
            formatted_message['title']  = message['message']

        if 'type' in message:
            formatted_message['type'] = message['type']

        if 'detail' in message:
            # TSL Fix
            if 'title' not in message and isinstance(message['detail'], str):
                formatted_message['title'] = message['detail']
            else:
                formatted_message['detail'] = message['detail']

        if 'status' in message:
            formatted_message['status'] = message['status']

        if 'instance' in message:
            formatted_message['instance'] = message['instance']

        return formatted_message

    @staticmethod
    def clean_message(message) -> str|dict:
        # Check if empty
        if not message:
            return ''

        # Check if Dictionary
        if isinstance(message, dict):
            return message

        # Check if JSON String
        try:
            return json.loads(message)
        except json.JSONDecodeError:
            return message

    @staticmethod
    def error(message=None, code=500):
        return APIRender.display(message, code)

    @staticmethod
    def get_plain_message(message=None, code=500) -> str:
        message = APIRender.clean_message(message)

        # RCP format 1
        if 'detail' in message:
            message = message['detail']

        # RCP format 2
        if 'message' in message:
            message = message['message']

        if not message:
            match code:
                case 401:
                    message = 'Unauthorized.'
                case _:
                    message = 'Unknown error.'

        return message

    @staticmethod
    def request_errors(e) -> JSONResponse:
        message = 'The given data was invalid.'
        formatted_errors = APIRender.clean_data(APIRender.process_errors(e))
        first_error = formatted_errors[0]
        if 'loc' in first_error and 'type' in first_error:
            loc = first_error['loc']
            type_ = first_error['type']
            message = f'Field {loc} is {type_}.' if loc != 'body' else f'Body is {type_}.'

        return APIRender.display({'title': message, 'detail': ''}, 422)

    @staticmethod
    def request_validation_handler(_: Request, e: ValidationError) -> JSONResponse:
        return APIRender.request_errors(e)

    @staticmethod
    def http_exception_handler(_: Request, e: HTTPException):
        message = APIRender.get_plain_message(e.detail, e.status_code)

        if e.status_code == HTTPStatus.UNAUTHORIZED:
            APIRender.display('Unauthorized.', HTTPStatus.UNAUTHORIZED)

        return APIRender.display(message, e.status_code)

    @staticmethod
    def app_exception_handler(_: Request, e: APIException):
        return APIRender.display(e.get_error(), e.code)

    @staticmethod
    def not_found(resource):
        message = f'{resource} not found.'
        return APIRender.display(message, 404)

    @staticmethod
    def loc_to_dot_sep(loc: tuple[str | int, ...]) -> str:
        path = ''
        for i, x in enumerate(loc):
            if isinstance(x, str):
                if i > 0:
                    path += '.'
                path += x
            elif isinstance(x, int):
                path += f'[{x}]'
            else:
                raise TypeError('Unexpected type')
        return path

    @staticmethod
    def process_errors(e: ValidationError) -> list[dict[str, Any]]:
        new_errors: list[dict[str, Any]] = e.errors()
        for error in new_errors:
            error['loc'] = APIRender.loc_to_dot_sep(error['loc'])
        return new_errors

    @staticmethod
    def clean_data(data):
        cleaned_data = []
        for item in data:
            cleaned_item = item.copy()
            if isinstance(cleaned_item.get('input'), type):
                cleaned_item['input'] = str(cleaned_item['input'])
            cleaned_data.append(cleaned_item)
        return cleaned_data


import math

from .query_builder import QueryBuilder
from .query_params import QueryParams


class Paginator:
    def __init__(self, query_builder: QueryBuilder, query_params: QueryParams):
        self.r = self.get_template()
        self.qp = query_params
        self.qb = query_builder

    @staticmethod
    def get_template() -> dict:
        return {
            'current_page': 1,
            'data': [],
            'total': None,
            'per_page': None,
            'total_pages': None
        }

    async def build(self):
        total = await self.qb.total()

        # Current Page
        self.r['current_page'] = self.qp.page

        # Data
        offset = self.qp.size * (self.qp.page - 1)
        self.qb.stmt = self.qb.stmt.offset(offset).limit(self.qp.size)
        self.r['data'] = await self.qb.get()

        # Total
        self.r['total'] = total

        # Per Page
        self.r['per_page'] = self.qp.size

        # Total Page
        self.r['total_pages'] = math.ceil(total / self.qp.size)

    async def get(self) -> dict:
        await self.build()
        return self.r
